from textwrap import dedent
from unittest import TestCase

from mock import patch

from jirafs.jirafieldmanager import JiraFieldManager


class TestJiraFieldManager(TestCase):
    def test_decode(self):
        encoded_values = dedent("""
            * summary:
                This is a test summary
            * longer_message:
                This is a much
                longer message that happens
                to contain newlines.
        """)

        expected_result = {
            'summary': 'This is a test summary',
            'longer_message': (
                'This is a much\n'
                'longer message that happens\n'
                'to contain newlines.'
            )
        }

        actual_result = JiraFieldManager(
            encoded_values
        )

        self.assertEqual(
            expected_result,
            actual_result
        )

    def test_decode_with_json_values(self):
        encoded_values = dedent("""
            * dictionary_field:
                {
                    "name": "Adam Coddington"
                }
            * integer_field:
                10
            * string_field:
                Hello
            * list_field:
                [
                    "Alphabet",
                    {
                        "One": "Two"
                    }
                ]
        """)

        expected_result = {
            "dictionary_field": {
                "name": "Adam Coddington"
            },
            "integer_field": 10,
            "string_field": "Hello",
            "list_field": [
                "Alphabet",
                {
                    "One": "Two",
                }
            ]
        }

        actual_result = JiraFieldManager(
            encoded_values
        )

        self.assertEqual(
            expected_result,
            actual_result,
        )

    def test_decode_with_human_readable_field_names(self):
        encoded_values = dedent("""
            * Summary:
                This is a test summary
            * Longer Message:
                This is a much
                longer message that happens
                to contain newlines.
            * Something(s) that are Company-Specific (customfield_108234):
                This is something else
        """)

        expected_result = {
            'Summary': 'This is a test summary',
            'Longer_Message': (
                'This is a much\n'
                'longer message that happens\n'
                'to contain newlines.'
            ),
            'customfield_108234': (
                'This is something else'
            )
        }

        actual_result = JiraFieldManager(
            encoded_values
        )

        self.assertEqual(
            expected_result,
            actual_result
        )




from __future__ import print_function

import json

from prettytable import PrettyTable

from jirafs import utils
from jirafs.plugin import CommandPlugin
from jirafs.ticketfolder import TicketFolder


class Command(CommandPlugin):
    """Search for users matching the specified search term"""
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    def handle(self, args, jira, path, parser, **kwargs):
        search_term = ' '.join(args.terms)

        try:
            tf = TicketFolder(path, jira)
            jira_client = tf.jira
        except IOError:
            jira_client = jira(
                utils.get_default_jira_server()
            )

        users = jira_client.search_users(search_term)

        if args.json:
            print(
                json.dumps(
                    [u.raw for u in users]
                )
            )
            return

        table = PrettyTable(['Name', 'Username', 'Email Address', 'Time Zone'])
        table.align = 'l'
        for user in users:
            table.add_row(
                [
                    user.displayName,
                    user.name,
                    user.emailAddress,
                    user.timeZone
                ]
            )
        print(table)

    def add_arguments(self, parser):
        parser.add_argument('terms', nargs='*')
        parser.add_argument('--json', default=False, action='store_true')



import six

from jirafs import utils
from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """Create a new JIRA issue"""
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    FIELDS = (
        {
            'name': 'project',
            'prompt': 'Project',
            'required': True,
            'path': 'project.key',
        },
        {
            'name': 'issuetype',
            'prompt': 'Issue Type',
            'default': 'Task',
            'path': 'issuetype.name',
        },
        {
            'name': 'summary',
            'prompt': 'Summary',
            'required': True,
        },
        {
            'name': 'description',
            'prompt': 'Description',
            'required': False,
        }
    )

    def set_field_value(self, data, field, value):
        starting_reference = data

        path = field.get('path', field['name'])
        for part in path.split('.'):
            if part not in data:
                data[part] = {}
            last_reference = data
            data = data[part]

        last_reference[path.split('.')[-1]] = value

        return starting_reference

    def prompt_for_input(self, field):
        while True:
            if field.get('default'):
                value = raw_input('%s (%s): ' % (
                    field.get('prompt'),
                    field.get('default'),
                ))
            else:
                value = raw_input('%s: ' % (
                    field.get('prompt')
                ))

            value = value.strip()

            if value:
                return value
            elif not field.get('required') and field.get('default'):
                return field.get('default')
            elif not field.get('required'):
                return value

    def handle(self, args, jira, path, parser, **kwargs):
        server = args.server
        if not server:
            server = utils.get_default_jira_server()

        issue_data = {}

        for field in self.FIELDS:
            if getattr(args, field['name']) is not None:
                self.set_field_value(
                    issue_data, field, getattr(args, field['name'])
                )
            elif args.quiet:
                self.set_field_value(issue_data, field, field.get('default'))
            else:
                self.set_field_value(
                    issue_data,
                    field,
                    self.prompt_for_input(field)
                )

        jira_client = jira(server)
        issue = jira_client.create_issue(issue_data)

        return run_command_method_with_kwargs(
            'clone',
            path=None,
            url=issue.permalink(),
            jira=jira,
        )

    def add_arguments(self, parser):
        parser.add_argument('--server', default=None)
        parser.add_argument(
            '--quiet', '-q', default=False, action='store_true'
        )
        for argument in self.FIELDS:
            parser.add_argument(
                '--%s' % argument['name'],
                default=None,
                type=six.text_type
            )



import argparse
import codecs
import logging
import os
import sys
import time

from blessings import Terminal
import six
from verlib import NormalizedVersion

from . import utils
from .exceptions import (
    GitCommandError,
    JiraInteractionFailed,
    JirafsError,
    NotTicketFolderException
)


# Write data to stdout as UTF-8 bytes when there's no encoding specified
if sys.version_info < (3, ) and sys.stdout.encoding is None:
    sys.stdout = codecs.getwriter('utf-8')(sys.stdout)


logger = logging.getLogger(__name__)


def main():
    term = Terminal()
    if sys.version_info < (2, 7):
        raise RuntimeError(
            "Jirafs requires minimally version 2.7 of Python 2, or "
            "any version of Python 3.  Please upgrade your version of "
            "python before using Jirafs."
        )
    if utils.get_git_version() < NormalizedVersion('1.8'):
        raise RuntimeError(
            "Jirafs requires minimally version 1.8 of Git.  Please "
            "upgrade your version of git before using Jirafs."
        )

    commands = utils.get_installed_commands()

    parser = argparse.ArgumentParser(
        description='Edit Jira issues locally from your filesystem',
        add_help=False,
    )
    parser.add_argument(
        'command',
        type=six.text_type,
        choices=commands.keys()
    )
    parser.add_argument(
        '--subtasks',
        action='store_true',
        default=False
    )
    parser.add_argument(
        '--log-level',
        default=None,
        dest='log_level',
    )
    args, extra = parser.parse_known_args()

    if args.log_level is not None:
        logging.basicConfig(level=logging.getLevelName(args.log_level))

    command_name = args.command
    cmd_class = commands[command_name]

    # Subtasks
    if args.subtasks:
        cmd_class.RUN_FOR_SUBTASKS = True

    started = time.time()
    logger.debug(
        'Command %s(%s) started',
        command_name,
        extra
    )
    jira = utils.lazy_get_jira()
    try:
        cmd_class.execute_command(
            extra, jira=jira, path=os.getcwd(), command_name=command_name
        )
    except GitCommandError as e:
        print(
            u"{t.red}Error (code: {code}) while running git "
            u"command.{t.normal}".format(
                t=term,
                code=e.returncode
            )
        )
        print("")
        print(u"{t.red}Command:{t.normal}{t.red}{t.bold}".format(t=term))
        print(u"    {cmd}".format(cmd=e.command))
        print(u"{t.normal}".format(t=term))
        print(u"{t.red}Output:{t.normal}{t.red}{t.bold}".format(t=term))
        for line in e.output.decode('utf8').split('\n'):
            print(u"    %s" % line)
        print(u"{t.normal}".format(t=term))
        sys.exit(10)
    except NotTicketFolderException:
        if not getattr(cmd_class, 'TRY_SUBFOLDERS', False):
            print(
                u"{t.red}The command '{cmd}' must be ran from "
                u"within an issue folder.{t.normal}".format(
                    t=term,
                    cmd=command_name
                )
            )
            sys.exit(20)
        count_runs = 0
        for folder in os.listdir(os.getcwd()):
            try:
                cmd_class.execute_command(
                    extra,
                    jira=jira,
                    path=os.path.join(
                        os.getcwd(),
                        folder,
                    ),
                    command_name=command_name,
                )
                count_runs += 1
            except NotTicketFolderException:
                pass
        if count_runs == 0:
            print(
                u"{t.red}The command '{cmd}' must be ran from "
                u"within an issue folder or from within a folder containing "
                u"issue folders.{t.normal}".format(
                    t=term,
                    cmd=command_name
                )
            )
            sys.exit(21)
    except JiraInteractionFailed as e:
        print(
            u"{t.red}JIRA was unable to satisfy your "
            u"request: {t.normal}{t.red}{t.bold}{error}{t.normal}".format(
                t=term,
                error=str(e)
            )
        )
        sys.exit(80)
    except JirafsError as e:
        print(
            u"{t.red}Jirafs encountered an error processing your "
            u"request: {t.normal}{t.red}{t.bold}{error}{t.normal}".format(
                t=term,
                error=str(e)
            )
        )
        sys.exit(90)

    logger.debug(
        'Command %s(%s) finished in %s seconds',
        command_name,
        extra,
        (time.time() - started)
    )



import io
import json
import os
import textwrap

import six

from jirafs import constants, utils
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Fetch remote changes """
    TRY_SUBFOLDERS = True
    RUN_FOR_SUBTASKS = False
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, folder, **kwargs):
        return self.fetch(folder)

    def get_field_map(self, folder):
        fields = {}
        for field in folder.jira.fields():
            fields[field['id']] = field['name']

        return fields

    def fetch(self, folder):
        folder.clear_cache()

        file_meta = folder.get_remote_file_metadata(shadow=True)
        original_hash = folder.run_git_command('rev-parse', 'jira')

        for filename in folder.get_remotely_changed():
            for attachment in folder.issue.fields.attachment:
                if attachment.filename == filename:
                    folder.log(
                        'Download file "%s"',
                        (attachment.filename, ),
                    )
                    content = six.BytesIO(attachment.get())
                    filename, content = folder.execute_plugin_method_series(
                        'alter_file_download',
                        args=((filename, content, ),),
                        single_response=True,
                    )
                    save_path = folder.get_shadow_path(filename)
                    with open(save_path, 'wb') as save_file:
                        content.seek(0)
                        save_file.write(content.read())
                        file_meta[filename] = attachment.created

        folder.set_remote_file_metadata(file_meta, shadow=True)

        field_map = self.get_field_map(folder)
        detail_path = folder.get_shadow_path(constants.TICKET_DETAILS)
        with io.open(detail_path, 'w', encoding='utf-8') as dets:
            for field in sorted(folder.issue.raw['fields'].keys()):
                value = folder.issue.raw['fields'][field]
                if isinstance(value, six.string_types):
                    value = value.replace('\r\n', '\n').strip()
                elif value is None:
                    value = ''
                elif field in constants.NO_DETAIL_FIELDS:
                    continue

                if not isinstance(value, six.string_types):
                    value = json.dumps(
                        value,
                        sort_keys=True,
                        indent=4,
                        ensure_ascii=False
                    )

                if field in constants.FILE_FIELDS:
                    # Write specific fields to their own files without
                    # significant alteration

                    file_field_path = folder.get_shadow_path(
                        constants.TICKET_FILE_FIELD_TEMPLATE
                    ).format(field_name=field)
                    with io.open(
                        file_field_path,
                        'w',
                        encoding='utf-8'
                    ) as file_field_file:
                        file_field_file.write(six.text_type(value))
                        file_field_file.write(
                            six.text_type('\n')
                        )  # For unix' sake
                else:
                    # Normal fields, though, just go into the standard
                    # fields file.
                    if value is None:
                        continue
                    elif field in constants.NO_DETAIL_FIELDS:
                        continue

                    human_readable = field_map.get(field)
                    if human_readable is None or human_readable == field:
                        dets.write(
                            six.text_type('* %s:\n') % (
                                field
                            )
                        )
                    else:
                        dets.write(
                            six.text_type('* %s (%s):\n') % (
                                human_readable,
                                field
                            )
                        )
                    for line in value.replace('\r\n', '\n').split('\n'):
                        dets.write(six.text_type('    %s\n' % line))

        links_path = folder.get_shadow_path(constants.TICKET_LINKS)
        with io.open(links_path, 'w', encoding='utf-8') as links_handle:
            # Write issue links
            for link in folder.issue.fields.issuelinks:
                category = 'outward'
                if 'inwardIssue' in link.raw:
                    category = 'inward'

                links_handle.write(
                    six.u("* {status}: {key}\n").format(
                        status=getattr(link.type, category).title(),
                        key=getattr(link, '%sIssue' % category).key
                    )
                )

            # Write remote links
            for link in folder.jira.remote_links(folder.issue):
                if link.object.title:
                    links_handle.write(
                        six.u("* {title}: {url}\n").format(
                            title=link.object.title,
                            url=link.object.url
                        )
                    )
                else:
                    links_handle.write(
                        six.u("* {url}\n").format(
                            title=link.object.title,
                            url=link.object.url
                        )
                    )

        comments_filename = folder.get_shadow_path(constants.TICKET_COMMENTS)
        with io.open(comments_filename, 'w', encoding='utf-8') as comm:
            for comment in folder.issue.fields.comment.comments:
                comm.write(
                    six.text_type('* At %s, %s wrote:\n\n') % (
                        comment.created,
                        comment.author
                    )
                )
                final_lines = []
                lines = comment.body.replace('\r\n', '\n').split('\n')
                for line in lines:
                    if not line:
                        final_lines.append(six.text_type(''))
                    else:
                        final_lines.extend(
                            textwrap.wrap(
                                line,
                                width=70,
                                expand_tabs=False,
                                replace_whitespace=False,
                                break_long_words=False,
                            )
                        )
                for line in final_lines:
                    comm.write(six.text_type('    %s\n') % line)
                comm.write(six.text_type('\n'))

        folder.store_cached_issue()

        # Clone subtasks
        subtasks = folder.issue.fields.subtasks
        if len(subtasks) > 0:
            commands = utils.get_installed_commands()
            jira = utils.lazy_get_jira()
            with open(
                folder.get_metadata_path('subtasks'),
                'w'
            ) as out:
                for issue in subtasks:
                    out.write(
                        '%s\n' % issue.key
                    )
                    issue_path = folder.get_path(issue.key)
                    if not os.path.exists(issue_path):
                        command_name = 'clone'
                        args = [
                            issue.permalink(),
                            issue_path
                        ]
                        path = folder.path
                        commands[command_name].execute_command(
                            args,
                            jira=jira,
                            path=path,
                            command_name=command_name
                        )
        folder.build_ignore_files()

        folder.run_git_command('add', '-A', shadow=True)
        folder.run_git_command(
            'commit', '-m', 'Fetched remote changes',
            failure_ok=True, shadow=True
        )
        folder.run_git_command('push', 'origin', 'jira', shadow=True)
        final_hash = folder.run_git_command('rev-parse', 'jira')
        if original_hash != final_hash:
            folder.log(
                "Updated 'jira' to %s" % final_hash
            )

        return utils.PostStatusResponse(
            original_hash == final_hash,
            final_hash
        )



import io
import json
import os
import shutil
import tempfile

import mock
from mock import patch
import six

from jirafs import exceptions
from jirafs.utils import run_command_method_with_kwargs
from jirafs.jirafieldmanager import JiraFieldManager

from .base import BaseTestCase


class TestTicketFolder(BaseTestCase):
    def setUp(self):
        self.arbitrary_ticket_number = 'ALPHA-123'
        self.root_folder = tempfile.mkdtemp()
        self.mock_jira = mock.MagicMock()
        self.mock_jira.issue.return_value = self.rehydrate_issue(
            'basic.issue.json'
        )
        self.mock_get_jira = lambda _, config=None: self.mock_jira

        with patch(
            'jirafs.ticketfolder.TicketFolder.get_remotely_changed'
        ) as get_remotely_changed:
            get_remotely_changed.return_value = []
            self.ticketfolder = run_command_method_with_kwargs(
                'clone',
                url='http://arbitrary.com/browse/ALPHA-123',
                jira=self.mock_get_jira,
                path=os.path.join(
                    self.root_folder,
                    self.arbitrary_ticket_number,
                )
            )

    def test_cloned_issue_successfully(self):
        paths = [
            'comments.read_only.jira',
            'fields.jira',
            'description.jira',
            'new_comment.jira',
            '.jirafs/issue.json',
            '.jirafs/remote_files.json',
            '.jirafs/version',
        ]
        for path in paths:
            self.assertTrue(
                os.path.isfile(
                    self.ticketfolder.get_local_path(path)
                ),
                '%s does not exist' % path
            )
            # Ensure that the file is tracked in git
            try:
                self.ticketfolder.run_git_command(
                    'ls-files', path, '--error-unmatch'
                )
            except exceptions.GitCommandError:
                self.fail(
                    "File %s is not tracked in the git repository."
                )

    def test_get_fields(self):
        actual_result = self.ticketfolder.get_fields()

        expected_result = json.loads(
            self.get_asset_contents('basic.status.json')
        )

        self.assertEquals(
            actual_result,
            expected_result
        )

    def test_fetch(self):
        self.ticketfolder._issue = (
            self.rehydrate_issue('test_fetch/fetched.json')
        )
        with patch.object(self.ticketfolder, 'clear_cache') as clear_cache:
            run_command_method_with_kwargs(
                'fetch',
                folder=self.ticketfolder,
            )
            self.assertTrue(clear_cache.called)

        expected_result = JiraFieldManager(
            self.get_asset_contents('test_fetch/fetched.jira')
        )
        with io.open(
            self.ticketfolder.get_shadow_path('fields.jira'),
            encoding='utf-8'
        ) as _in:
            actual_result = JiraFieldManager(_in.read())

        self.assertEqual(actual_result, expected_result)

    def test_push(self):
        changed_field = 'description'
        changed_value = 'Something Else'

        status = self.get_empty_status()
        status['ready']['fields'][changed_field] = (
            'Something', changed_value,
        )
        with patch('jirafs.commands.pull.Command.pull') as pull:
            with patch.object(self.ticketfolder, 'status') as status_method:
                status_method.return_value = status
                with patch.object(self.ticketfolder.issue, 'update') as out:
                    run_command_method_with_kwargs(
                        'push',
                        folder=self.ticketfolder
                    )
                    self.assertTrue(pull.called)
                    out.assert_called_with(
                        **{
                            'description': changed_value
                        }
                    )

    def test_push_rejected_if_updated(self):
        src_path = self.get_asset_path('test_fetch/fetched.jira')
        dst_path = self.ticketfolder.get_shadow_path('fields.jira')
        shutil.copyfile(
            src_path,
            dst_path,
        )

        self.ticketfolder.run_git_command('add', '-A', shadow=True)
        self.ticketfolder.run_git_command(
            'commit', '-m', 'Changed', shadow=True
        )
        self.ticketfolder.run_git_command(
            'push', 'origin', 'jira', shadow=True
        )
        with self.assertRaises(exceptions.LocalCopyOutOfDate):
            run_command_method_with_kwargs(
                'push',
                folder=self.ticketfolder
            )

    def test_status_new_file(self):
        src_path = self.get_asset_path('test_status_local_changes/alpha.svg')
        dst_path = self.ticketfolder.get_local_path('alpha.svg')
        shutil.copyfile(
            src_path,
            dst_path,
        )

        expected_output = self.get_empty_status()
        expected_output['uncommitted']['files'] = ['alpha.svg']
        actual_output = self.ticketfolder.status()

        self.assertEqual(expected_output, actual_output)

    def test_status_local_changes(self):
        src_path = self.get_asset_path('test_fetch/fetched.jira')
        dst_path = self.ticketfolder.get_local_path('fields.jira')
        shutil.copyfile(
            src_path,
            dst_path,
        )

        expected_output = self.get_empty_status()
        expected_output['uncommitted']['fields']['assignee'] = (
            '',
            {
                "active": True,
                "displayName": (
                    u"Coddington, Adam (\u0410\u0440\u0431\u0438\u0442"
                    u"\u0440\u0430\u0440\u0438\u041a\u043e\u0440\u043f"
                    u"-Atlantis)"
                ),
                "name": "acoddington",
            }
        )
        actual_output = self.ticketfolder.status()

        for field in expected_output:
            if expected_output[field] != actual_output[field]:
                self.fail(
                    "Field %s does not match; Actual: %s != Expected: %s" % (
                        field,
                        actual_output[field],
                        expected_output[field],
                    )
                )

        self.assertEqual(expected_output, actual_output)

    def test_status_new_comment(self):
        arbitrary_comment = six.text_type('New Comment')

        comment = self.ticketfolder.get_local_path('new_comment.jira')
        with io.open(comment, 'w', encoding='utf-8') as out:
            out.write(arbitrary_comment)

        expected_output = self.get_empty_status()
        expected_output['uncommitted']['new_comment'] = arbitrary_comment
        actual_output = self.ticketfolder.status()

        self.assertEqual(expected_output, actual_output)

    def tearDown(self):
        shutil.rmtree(self.root_folder)



from jirafs.exceptions import JirafsError
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Get the status of the current ticketfolder """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.field(folder, args.field_name)

    def add_arguments(self, parser):
        parser.add_argument(
            'field_name',
        )

    def field(self, folder, field_name):
        fields = folder.get_fields()

        if field_name not in fields:
            raise JirafsError("Field '%s' does not exist." % field_name)

        print(fields[field_name])



from blessings import Terminal
import six

from jirafs import utils
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Enable/Disable or display information about installed issue plugins """
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, parser, **kwargs):
        installed_plugins = utils.get_installed_plugins()
        if args.disabled_only and args.enabled_only:
            parser.error(
                "--disabled-only and --enabled-only are mutually exclusive."
            )
        if args.enable and args.enable not in installed_plugins:
            parser.error(
                "Plugin '%s' is not installed." % args.enable
            )

        return self.plugins(folder, args)

    def add_arguments(self, parser):
        parser.add_argument('--verbose', action='store_true', default=False)
        parser.add_argument(
            '--enabled-only',
            dest='enabled_only',
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--disabled-only',
            dest='disabled_only',
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--enable',
            type=six.u,
        )
        parser.add_argument(
            '--disable',
            type=six.u,
        )
        parser.add_argument(
            '--global',
            dest='set_global',
            default=False,
            action='store_true',
        )

    def build_plugin_dict(self, enabled, available):
        all_plugins = {}
        for plugin_name, cls in available.items():
            all_plugins[plugin_name] = {
                'enabled': False,
                'class': cls,
            }
        for plugin_instance in enabled:
            plugin_name = plugin_instance.plugin_name
            all_plugins[plugin_name]['enabled'] = True
            all_plugins[plugin_name]['instance'] = plugin_instance

        return all_plugins

    def plugins(self, folder, args):
        t = Terminal()

        enabled_plugins = folder.load_plugins()
        available_plugins = utils.get_installed_plugins()

        if args.enable:
            if args.set_global:
                utils.set_global_config_value(
                    'plugins', args.enable, 'enabled',
                )
            else:
                folder.set_config_value(
                    'plugins', args.enable, 'enabled'
                )
        elif args.disable:
            if args.set_global:
                utils.set_global_config_value(
                    'plugins', args.disable, 'disabled',
                )
            else:
                folder.set_config_value(
                    'plugins', args.disable, 'disabled'
                )
        else:
            all_plugins = self.build_plugin_dict(
                enabled_plugins,
                available_plugins
            )

            for plugin_name, plugin_data in all_plugins.items():
                if plugin_data['enabled'] and args.disabled_only:
                    continue
                if not plugin_data['enabled'] and args.enabled_only:
                    continue
                if plugin_data['enabled']:
                    color = t.bold
                else:
                    color = t.normal

                print(
                    color + plugin_name + t.normal +
                    (
                        ' (Enabled)'
                        if plugin_data['enabled']
                        else (
                            ' (Disabled; enable by running `jirafs '
                            'plugins --enable=%s`)' % plugin_name
                        )
                    )
                )
                if args.verbose:
                    doc_string = (
                        plugin_data['class'].__doc__.strip().split('\n')
                    )
                    for line in doc_string:
                        print('     %s' % line)



import json

from blessings import Terminal

from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Get the status of the current ticketfolder """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.status(folder, args.format)

    def add_arguments(self, parser):
        parser.add_argument(
            '--format',
            default='text',
            choices=['text', 'json']
        )

    def status(self, folder, output_format='text'):
        status = folder.status()
        if output_format == 'json':
            self.status_json(folder, status)
        self.status_text(folder, status)
        return status

    def status_json(self, folder, status):
        print(json.dumps(status))

    def has_changes(self, section, *keys):
        if not keys:
            keys = ['files', 'fields', 'new_comment', 'links']
        for key in keys:
            if section[key]:
                return True

    def status_text(self, folder, folder_status):
        t = Terminal()
        print(
            u"On ticket {ticket} ({url})".format(
                ticket=folder.ticket_number,
                url=folder.cached_issue.permalink(),
            )
        )
        if not folder_status['up_to_date']:
            print(
                t.magenta + "Warning: unmerged upstream changes exist; "
                "run `jirafs merge` to merge them into your local copy." +
                t.normal
            )

        printed_changes = False
        ready = folder_status['ready']
        if self.has_changes(ready):
            printed_changes = True
            print('')
            print(
                "Ready for upload; use `jirafs push` to update JIRA."
            )
            print(
                self.format_field_changes(ready, 'green', terminal=t)
            )

        staged = folder_status['uncommitted']
        if self.has_changes(staged):
            printed_changes = True
            print('')
            print(
                "Uncommitted changes; use `jirafs commit` to mark these "
                "for JIRA."
            )
            print(
                self.format_field_changes(staged, 'red', terminal=t)
            )

        local_uncommitted = folder_status['local_uncommitted']
        if self.has_changes(local_uncommitted, 'files'):
            printed_changes = True
            print('')
            print(
                "Uncommitted changes prevented from being sent to JIRA "
                "because they match at least one of the patterns in your "
                ".jirafs_ignore file; use `jirafs commit` to commit these "
                "changes."
            )
            print(
                "Note: these files will " + t.bold + "not" + t.normal + " "
                "be uploaded to JIRA even after being committed."
            )
            print(
                self.format_field_changes(
                    local_uncommitted,
                    'cyan',
                    no_upload=True,
                    terminal=t,
                )
            )

        if not printed_changes:
            print('No changes found')
        else:
            print("")
            print('For more detail about these changes, run `jirafs diff`')

    def format_field_changes(
        self, changes, color, no_upload=False, terminal=None
    ):
        if terminal is None:
            t = Terminal()
        else:
            t = terminal
        lines = []
        color = getattr(t, color)
        normal = t.normal

        for filename in changes.get('files', []):
            lines.append(
                '\t' + color + filename + normal + (
                    ' (save to repository)' if no_upload else ' (file upload)'
                )
            )
        for link, data in changes.get('links', {}).get('remote', {}).items():
            orig = data[0]
            new = data[1]
            if new is not None:
                if new.get('description'):
                    description = new['description']
                else:
                    description = '(Untitled)'
                lines.append(
                    '\t' + color + description +
                    ': ' + link + normal + (
                        ' (changed remote link)'
                        if orig else ' (new remote link)'
                    )
                )
            else:
                if orig.get('description'):
                    description = orig['description']
                else:
                    description = '(Untitled)'
                lines.append(
                    '\t' + color + description +
                    ': ' + link + normal + ' (removed remote link)'
                )
        for link, data in changes.get('links', {}).get('issue', {}).items():
            orig = data[0]
            new = data[1]
            if new is not None:
                if new.get('status'):
                    status = new['status']
                else:
                    status = '(Untitled)'
                lines.append(
                    '\t' + color + status.title() +
                    ': ' + link + normal + (
                        ' (changed issue link)'
                        if orig else ' (new issue link)'
                    )
                )
            else:
                if orig.get('status'):
                    status = orig['status']
                else:
                    status = '(Untitled)'
                lines.append(
                    '\t' + color + status.title() +
                    ': ' + link + normal + ' (removed issue link)'
                )
        for field, value_set in changes.get('fields', {}).items():
            lines.append(
                '\t' + color + field + normal
            )
        if changes.get('new_comment', ''):
            lines.append(
                '\t' + color + '[New Comment]' + normal
            )

        return '\n'.join(lines)



import webbrowser

from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Open the current ticketfolder's issue in your web browser """
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, folder, **kwargs):
        return self.open(folder)

    def open(self, folder):
        return webbrowser.open(folder.cached_issue.permalink())



from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Print a diff of locally-changed files """
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, folder, **kwargs):
        return self.diff(folder)

    def diff(self, folder):
        result = folder.run_git_command('diff')
        if result:
            result = result.strip()
        print(result)
        return result



from jirafs import exceptions, utils
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Commit local changes for later submission to JIRA """
    MIN_VERSION = '1.0a1'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.commit(folder, args.message, *args.git_arguments)

    def add_arguments(self, parser):
        parser.add_argument(
            '-m', '--message', dest='message', default='Untitled'
        )
        parser.add_argument(
            'git_arguments', nargs='*'
        )

    def commit(self, folder, message, *args):
        original_hash = folder.run_git_command('rev-parse', 'master')
        folder.run_git_command(
            'add', '-A'
        )
        try:
            folder.run_git_command(
                'commit', '-m', message, *args
            )
            final_hash = folder.run_git_command('rev-parse', 'master')
            folder.log(
                "Changes committed; current hash is %s" % final_hash
            )
        except exceptions.GitCommandError:
            print("Nothing to commit")
        final_hash = folder.run_git_command('rev-parse', 'master')
        return utils.PostStatusResponse(
            original_hash == final_hash,
            final_hash
        )
